ccc  kapgnfwrvir_zl_ns_de.f   Time-stamp: <2007-07-25 12:03:02 hamana>
ccc  
ccc  This computes the redshift distribution of WL selected clusters.
ccc  This allows to compute it for the NFW halo (alpha=1) truncated 
ccc  at r=rvir. 
ccc  This is exactly  the program used to compute the counts presented in 
ccc  Hamana, Takada & Yoshida 2004, MNRAS, 350, 893.
ccc  
ccc  This is for an empirically modeled source redshift distribution.
ccc  Notice this is VALID only for FLAT model
ccc  
ccc  If this does not work well, try a compile option that defines 
ccc  REAL declarations as DOUBLE PRECISION such like; 
ccc  "-real_size 64" (alpha), "-r8" (alpha)
ccc
ccc  This is a FREE fortran90 program
ccc  written by Takashi Hamana 2007/7/25
ccc  
      implicit real (a-h,o-z)
      implicit integer (i-n)
!!!
      common /cosmo/ om,ode,w0,wa
!!!
      common /omegas/ omem,omev,omek
      common /matter_Pk/ gamfin,atps
      common /defs/ pi,tps,ch0,rn,onethr
      common /st/ p_st,a_st,st_norm
      common /conc/ b0,c0,zpow
      common /tgauss/ r_gauss,czl,czl1

      external akap
      external dldz
!!!
      external dw
!!!

      include 'kapgnfwrvir_zl_ns_de.par'
      
!!!
ccc growth factor
      external derivs
      parameter (nstep=1000)
      parameter (hrk=0.001)
      dimension dplus(0:nstep)
      dimension growpd(0:nstep)
      parameter (nrk4=2)
      real yrk(nrk4),dydxrk(nrk4),youtrk(nrk4)
!!!

      dimension ds(inta)

      character*88 filename

ccc useful numbers
      pi=4.0*atan(1.0)
      arcmin2rad=pi/(180.0*60.0)
      domega_1sqdeg=tan(pi/180.0)**2
      tps=2.0*pi**2
      ch0=3000.0
      rn=8.0/ch0
      onethr=1.0/3.0
      rhocr=2.78e11 !h^2 M_sun/Mpc^3
      e2ten=2.30259 !ln(x)=ln(10)*log(x), ln(10)=e2ten
      zero=0.0

ccc flags
      igrfrag=-1

ccc mass function
      p_st=0.3  ! ST parameter 
      a_st=0.707  ! ST parameter 

ccc parameter input 
!!!
ccc cosmology for the normalization
! Hubble, Omega_m, Omega_lambda, Omega_b*hubp**2, sigma_8
      read(*,*)hubp,omem,omev,omebh2,sig8
      om=omem
      ode=omev
      w0=-1.0
      wa=0.0
      omeb=omebh2/hubp**2
      omek=omem+omev-1.0

      hubp0=hubp
      omem0=omem
      omev0=omev
      omeb0=omeb
      sig80=sig8
ccc Gamma factor
      gamfin=exp(omeb+sqrt(2.0*hubp)*omeb/omem)/
     $     (omem*hubp*ch0)
ccc normalization
      call normPk(anrm,sig8)
      atps=anrm/tps

ccc compute the linear growth factor D+(a)=a*g(a)
      ax=0.0
      growpd(0)=1.0

      x=ax
      yrk(1)=1.0
      yrk(2)=0.0
      dydxrk(1)=0.0
      dydxrk(2)=0.0
      call rk4(yrk,dydxrk,nrk4,x,hrk,youtrk,derivs)
      yrk(1)=youtrk(1)
      yrk(2)=youtrk(2)
      growpd(1)=yrk(1)
      
      do i=2,nstep
         x=hrk*(i-1)
         call derivs(x,yrk,dydxrk)
         call rk4(yrk,dydxrk,nrk4,x,hrk,youtrk,derivs)
         yrk(1)=youtrk(1)
         yrk(2)=youtrk(2)
         growpd(i)=yrk(1)
      enddo
      growpd_norm=growpd(nstep)

ccc cosmological model
! Hubble, Omega_m, Omega_de, w0, wa, Omega_b*hubp**2, sigma_8
      read(*,*)hubp,omem,ode,w0,wa,omebh2
      om=omem
      omeb=omebh2/hubp**2
      omek=0.0
ccc Gamma factor
      gamfin=exp(omeb+sqrt(2.0*hubp)*omeb/omem)/
     $     (omem*hubp*ch0)
      
ccc compute the linear growth factor D+(a)=a*g(a)
      ax=0.0
      growpd(0)=1.0

      x=ax
      yrk(1)=1.0
      yrk(2)=0.0
      dydxrk(1)=0.0
      dydxrk(2)=0.0
      call rk4(yrk,dydxrk,nrk4,x,hrk,youtrk,derivs)
      yrk(1)=youtrk(1)
      yrk(2)=youtrk(2)
      growpd(1)=yrk(1)
      
      do i=2,nstep
         x=hrk*(i-1)
         call derivs(x,yrk,dydxrk)
         call rk4(yrk,dydxrk,nrk4,x,hrk,youtrk,derivs)
         yrk(1)=youtrk(1)
         yrk(2)=youtrk(2)
         growpd(i)=yrk(1)
      enddo
      do i=0,nstep
         ax=i*hrk
         dplus(i)=growpd(i)/growpd(nstep)*ax
      enddo

ccc scaling sigma_8
c      write(*,*)sig8,anrm,atps
c      write(*,*)growpd(nstep),growpd_norm
      sig8=sig8*growpd(nstep)/growpd_norm
      anrm=anrm*(growpd(nstep)/growpd_norm)**2
      atps=atps*(growpd(nstep)/growpd_norm)**2
c      write(*,*)sig8,anrm,atps

ccc precompute coefficients
      coefdnu=e2ten*delnu*0.75/pi*ch0**3*domega_1sqdeg
      r2mass=2.775e11*4.0*pi*omem/3.0
!!!
      
ccc concentration
      read(*,*)c0,b0,zpow
      zpow=-zpow

ccc zmax
      zsmax=5.0
      asmax=1.0/(1.0+zsmax)
      delas=(1.0-asmax)/inta

      do izs=1,inta
         as=1.0-izs*delas
!!!
         call qromb(dw,as,1.0,ss)
!!!
         ds(izs)=ss
      enddo

ccc ns(z): redshift distribution
      read(*,*)alpns,betans,zmean
      xx=(1.0+alpns)/betans
      g1=exp(gammln(xx))
      xx=(2.0+alpns)/betans
      g2=exp(gammln(xx))
      zast=zmean*g1/g2
      coefns=betans/(zast*g1)
      
ccc noise 
      read(*,*)sigmae,gnum,theta_gauss  ! in arcmin
      sigmaero=0.5*sigmae/(sqrt(pi*gnum)*theta_gauss)
      
ccc SN lebel
      snlebel3=3.0
      snlebel4=4.0
      snlebel5=5.0
      akapmin3=snlebel3*sigmaero
      akapmin4=snlebel4*sigmaero
      akapmin5=snlebel5*sigmaero
            
ccc output file name
      read(*,'(a88)')filename

ccc pre-compute ST mass function
      call int_st(st_norm)

ccc pre-compute sigma-R relation at z=0
      call int_sigmar

clear array
      count3=0.0
      count4=0.0
      count5=0.0
      
C===== below thie line quantities depend on the lens redshift =====
ccc al(zl)->chi
      open(10,file=filename,STATUS= 'UNKNOWN')
      do 31 iz=1,inta-1 ! loop for lens redshift
         al=1.0-iz*delas
!!!
         call qromb(dw,al,1.0,ss)
!!!
         dl=ss
!!!
         dchida=dw(al)
!!!
         fchi=delas*dchida*dl**2

         ddeff=0.0
         do izs=iz+1,inta ! loop for zs
            as=1.0-izs*delas
            z=1.0/as-1.0
!!!
            call qromb(dw,as,al,ss)
!!!
            dd=ss/ds(izs)/as**2
            ddeff=ddeff+
     $           delas*dd*dndz(alpns,betans,zast,z)
         enddo
         ddeff=ddeff*coefns

ccc growth rate at zl
!!!
         ga2=gra(al,nstep,hrk,dplus)**2
         g=sqrt(ga2)/al
!!!

ccc Virial overdensity delta_vir at zl
         call deltas(al,deltac,deltavir)
         deltac2=deltac**2
         deltavir3=deltavir**onethr

ccc r_ast at zl
         call search_rast(rast,ga2,deltac2)

C===== below this line quantities depends also on the halo mass =====
ccc clear array
         dcount3=0.0
         dcount4=0.0
         dcount5=0.0
         do 30 inu=numin,numax  ! loop for nu
cccccc for nu=nu(inu))
            anu=10.0**(inu*delnu) ! nu
            sigma2_nu=deltac2/anu ! sigma^2
            rnu=sigma2r(sigma2_nu,ga2) ! R
            dn=st_norm*fnu_nu(anu)
ccc Virial radius and r_lin(M)
            rlin=rnu
            rvir=rlin/deltavir3
            flnnu=coefdnu*dn/rlin**3
ccc virial mass
            hmass=r2mass*rnu**3
            if (hmass.lt.1.0.or.hmass.gt.1.0e18) goto 55
ccc concentration parameter c[z,M]
            czl=amss2c(al,rlin,rast)
            if (czl.le.1.0) czl=1.0
            czl1=czl+1.0
ccc r_s
            rs=rvir/czl
ccc delta_s
            delta_s=deltavir*czl**3/
     $           (3.0*(log(czl1)-czl/czl1))
ccc kappa_s
            skappa=1.5*omem*dl*ddeff*delta_s*rs/(al*ch0)
ccc smoothing length in unit of r_s
            r_gauss=tan(theta_gauss*arcmin2rad)*dl*ch0/rs
ccc smoothed kappa
            call qromb(akap,0.0,5.0*r_gauss,ss1)
            thkap=skappa*4.0*ss1/r_gauss**2
            if (thkap.ge.akapmin3) dcount3=dcount3+flnnu*0.5
            if (thkap.ge.akapmin4) dcount4=dcount4+flnnu*0.5
            if (thkap.ge.akapmin5) dcount5=dcount5+flnnu*0.5
            
cccccc for nu=(nu(inu-1)+nu(inu))/2
 55         anu=10.0**((inu-0.5)*delnu) ! nu 
            sigma2_nu=deltac2/anu ! sigma^2
            rnu=sigma2r(sigma2_nu,ga2) ! R
            dn=st_norm*fnu_nu(anu)
ccc Virial radius and r_lin(M)
            rlin=rnu
            rvir=rlin/deltavir3
            flnnu=coefdnu*dn/rlin**3
ccc virial mass
            hmass=r2mass*rnu**3
            if (hmass.lt.1.0.or.hmass.gt.1.0e18) goto 30
ccc concentration parameter c[z,M]
            czl=amss2c(al,rlin,rast)
            if (czl.le.1.0) czl=1.0
            czl1=czl+1.0
ccc r_s
            rs=rvir/czl
ccc delta_s
            delta_s=deltavir*czl**3/
     $           (3.0*(log(czl1)-czl/czl1))
ccc kappa_s
            skappa=1.5*omem*dl*ddeff*delta_s*rs/(al*ch0)
ccc smoothing length in unit of r_s
            r_gauss=tan(theta_gauss*arcmin2rad)*dl*ch0/rs
ccc smoothed kappa
            call qromb(akap,0.0,5.0*r_gauss,ss1)
            thkap=skappa*4.0*ss1/r_gauss**2
            if (thkap.ge.akapmin3) dcount3=dcount3+flnnu*0.5
            if (thkap.ge.akapmin4) dcount4=dcount4+flnnu*0.5
            if (thkap.ge.akapmin5) dcount5=dcount5+flnnu*0.5
            
 30      continue
         dcount3=dcount3*fchi
         if (dcount3.le.1.e-33) dcount3=1.e-33
         dcount4=dcount4*fchi
         if (dcount4.le.1.e-33) dcount4=1.e-33
         dcount5=dcount5*fchi
         if (dcount5.le.1.e-33) dcount5=1.e-33
         count3=count3+dcount3
         count4=count4+dcount4
         count5=count5+dcount5
         redshift=1.0/al-1.0
         dndz3=dcount3*al**2/delas
         dndz4=dcount4*al**2/delas
         dndz5=dcount5*al**2/delas
         write(10,100)redshift,dndz3,count3,dndz4,count4,dndz5,count5
 31   continue
      
 100  format(7(1pe13.5))

      write(10,101)'# z; dN(>S/N)/dz; N(<z) #################'
      write(10,101)'# col2-3 : SN>3                          '
      write(10,101)'# col4-5 : SN>4                          '
      write(10,101)'# col6-7 : SN>5                          '
      write(10,102)'# N(>S/N=3)   =',count3
      write(10,102)'# N(>S/N=4)   =',count4
      write(10,102)'# N(>S/N=5)   =',count5
      write(10,101)'### cosmology ###########################'
      write(10,102)'# Hubble param=',hubp
      write(10,102)'# Omega_M     =',omem
!!!
      write(10,102)'# Omega_de    =',ode
      write(10,102)'# w0          =',w0
      write(10,102)'# wa          =',wa
!!!
      write(10,102)'# Omega_b     =',omeb
      write(10,102)'# sigma_8     =',sig8
      if (itkflag.eq.1)write(10,103)'# CDM TF = BBKS    '
      if (itkflag.eq.2)write(10,103)'# CDM TF = BE      '
      if (itkflag.eq.3)write(10,103)'# CDM TF = BBKS+S95'
!!!
      write(10,101)'### cosmology for normalization #########'
      write(10,102)'# Hubble param=',hubp0
      write(10,102)'# Omega_M     =',omem0
      write(10,102)'# Omega_lambda=',omev0
      write(10,102)'# Omega_b     =',omeb0
      write(10,102)'# sigma_8     =',sig80
!!!
      write(10,101)'### Halo parameters #####################'
      write(10,102)'# c_*         =',c0
      write(10,102)'# b0          =',b0
      write(10,102)'# zpow        =',zpow
      write(10,101)'# where c(z,M)=c0*(1+z)^zpow*(M/M14)^b0  '
      write(10,101)'### source distribution #################'
      write(10,102)'# zmean       =',zmean
      write(10,102)'# alpha       =',alpns
      write(10,102)'# beta        =',betans
      write(10,102)'# zast        =',zast
      write(10,101)'# where ns(z) propto                     '
      write(10,101)'#    (z/zast)**alpha exp(-(z/zast)**beta)'
      write(10,101)'### obs params ##########################'
      write(10,102)'# sigma_e     =',sigmae
      write(10,102)'# source dens =',gnum
      write(10,101)'### Gaussian kernel #####################'
      write(10,102)'# theta_G     =',theta_gauss
      close(10)
 101  format(a41)
 102  format(a15,1pe12.5)
 103  format(a19)
      end

      function dndz(alp,beta,zast,z)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      dndz=(z/zast)**alp*exp(-(z/zast)**beta)
      return
      end
      
      function fkap(y)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      common /tgauss/ r_gauss,c,c1

      fkap=0.0
      if (y.lt.1.0) then
         z=(y**2+c)/(y*c1)
         x=1.0-y**2
         fkap=-sqrt(c**2-y**2)/(x*c1)+log(z+sqrt(z**2-1.0))/x**1.5
      elseif (y.eq.1.0) then
         fkap=sqrt(c**2-1.0)*(1.0+1.0/c1)/(3.0*c1)
      elseif (y.gt.1.0.and.y.le.c) then
         z=(y**2+c)/(y*c1)
         x=1.0-y**2
         fkap=-sqrt(c**2-y**2)/(x*c1)-acos(z)/(-x)**1.5
      elseif (y.gt.c) then
         fkap=0.0
      endif
      return
      end


      function akap(y)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      common /tgauss/ r_gauss,c,c1
      z=(y/r_gauss)**2
      s=exp(-z)
      if (y.eq.0.0) then
         akap=0.0
      else
         akap=fkap(y)*y*s
      endif
      return
      end

ccc Sheth & Tormen mass function ! NOT normalized
      function fnu_nu(x)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      common /st/ p,a,st_norm
      y=a*x
      fnu_nu=(1.0+1.0/y**p)*sqrt(y)*exp(-0.5*y)
      return
      end
ccc normalization of Sheth & Tormen mass function    
      subroutine int_st(st_norm)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      external fnu_nu_log
      amin=-80.0
      amax=80.0
      call qromb(fnu_nu_log,amin,amax,ss)
      st_norm=1.0/ss
      return
      end
      function fnu_nu_log(z)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      common /st/ p,a,st_norm
      x=exp(z)
      y=a*x
      fnu_nu_log=(1.0+y**(-p))*sqrt(y)*exp(-0.5*y)
      return
      end

      function amss2c(as,r,rast)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      common /omegas/ omem,omev,omek
      common /defs/ pi,tps,ch0,rn,onethr
      common /conc/ b0,c0,zpow
ccc Bullock et al model
!!! 2.775e11/1e14=2.775e-3
      amass=2.775e-3*4.0*pi*omem*r**3/3.0
      amss2c=c0*as**zpow*amass**b0
      return
      end

ccc delta_c and delta_vir 
ccc VALID only FLAT model
      subroutine deltas(a,deltac,deltavir)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      common /omegas/ omem,omev,omek
      common /defs/ pi,tps,ch0,rn,onethr
      x=a*(1.0/omem-1.0)**(1.0/3.0)
      deltac=3.0*(12.0*pi)**(2.0/3.0)/20.0*
     $     (1.0-0.00123*log10(1.0+x**3))
      deltavir=18.0*pi**2*(1.0+0.4093*x**2.71572)
      return
      end


      subroutine search_rast(rast,ga2,deltac2)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      common /sigma_r_relation/ sigmar0,r0,drdsig1,drdsig2
      include 'kapgnfwrvir_zl_ns_de.par'
      dimension sigmar0(irall)
      dimension r0(irall)
      sig0=log(deltac2/ga2)
      call hunt(sigmar0,irall,sig0,jlo)
      d=(r0(jlo+1)-r0(jlo))/(sigmar0(jlo+1)-sigmar0(jlo))
      r1=d*(sig0-sigmar0(jlo))+r0(jlo)
      rast=exp(r1)
      return
      end
ccc sigma_R - R relation
      subroutine int_sigmar
      implicit real (a-h,o-z)
      implicit integer (i-n)
      common /sigma_r_relation/ sigmar0,r0,drdsig1,drdsig2
      include 'kapgnfwrvir_zl_ns_de.par'
      dimension sigmar0(irall)
      dimension r0(irall)
      do 10 i=1,irall
         ii=irmax-i+1
         x=10.0**(delr*ii)
         r0(i)=log(x)
         sigmar0(i)=log(sigma_r(x)) ! Notice sigma_r(x)=sigma_R^2
 10   continue
      drdsig1=(r0(2)-r0(1))/(sigmar0(2)-sigmar0(1))
      drdsig2=(r0(irall)-r0(irall-1))/
     $     (sigmar0(irall)-sigmar0(irall-1))
      return
      end
ccc
      function sigma2r(sig,ga2)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      common /sigma_r_relation/ sigmar0,r0,drdsig1,drdsig2
      include 'kapgnfwrvir_zl_ns_de.par'
      dimension sigmar0(irall)
      dimension r0(irall)
      sig0=log(sig/ga2)
      if (sig0.le.sigmar0(1)) then
         y=r0(1)+drdsig1*(sig0-sigmar0(1))
      elseif (sig0.ge.sigmar0(irall)) then
         y=r0(irall)+drdsig2*(sig0-sigmar0(irall))
      else
         call hunt(sigmar0,irall,sig0,jlo)
         d=(r0(jlo+1)-r0(jlo))/(sigmar0(jlo+1)-sigmar0(jlo))
         y=d*(sig0-sigmar0(jlo))+r0(jlo)
      endif
      sigma2r=exp(y)
      return
      end


ccccccccccccccccccccccccccccccccccccccccccccc
      function sigma_r(x)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      common /defs/ pi,tps,ch0,rn,onethr
      common /sigmar/rsmooth
      external dsigma_r
      wmin=-30.0
      wmax=40.0
      rsmooth=x/ch0
      call qromb(dsigma_r,wmin,wmax,ss)
      sigma_r=ss
      return
      end   
ccccccccccccccccccccccccccccccccccccccccccccc
      function dsigma_r(x)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      common /sigmar/rsmooth
      w=exp(x)
      rk=rsmooth*w
      dsigma_r=delta_lin0(w)*filtTH(rk)**2
      return
      end 
ccccccccccccccccccccccccccccccccccccccccccccc
      function delta_lin0(w)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      common /matter_Pk/ gfin,atps,itkflag
      q=w*gfin
      delta_lin0=atps*trnf2(q)*w**4
      return
      end 

cccccccccc normalization of power spectrum 
      subroutine normPk(a_norm,sig8)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      external delta
      common /defs/ pi,tps,ch0,rn,onethr
      wmin=-30.0
      wmax=40.0
      call qromb(delta,wmin,wmax,seki)
      y=seki/tps
      a_norm=sig8*sig8/y
      return
      end
ccccccccc
      function delta(x)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      common /defs/ pi,tps,ch0,rn,onethr
      common /matter_Pk/ gfin,atps,itkflag
      w=exp(x)
      q=w*gfin
      r=w*rn
      filt=filtTH(r)
      tranf=trnf2(q)
      delta=w**4*filt*filt*tranf
      return
      end
ccccccccc
      function filtTH(r)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      filtTH=3.0*(sin(r)-r*cos(r))/r**3
      return
      end
ccccccccc 
      function trnf2(q)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      common /matter_Pk/ gamfin,atps
      qq=q*q
      trnf2=0.182628*log(1.0+2.34*q)**2/qq
     $     /sqrt(1.0+3.89*q+259.21*qq
     $     +162.77*qq*q+2027.17*qq*qq)
      return
      end
ccccccccccccccccccccccccccccccccccccccccccccccccccc
      function gr(a,igrfrag)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      common /omegas/ omem,omev,omek
      save grnorm
      if (igrfrag.eq.-1) then
         pre=1.0
         et=1.0/e(pre)
         ot=omem*et
         vt=omev*et
         grnorm=0.4*(ot**(0.571428571)-vt+
     $        (1.0+0.5*ot)*(1.0+vt*0.014285714))/ot
         igrfrag=1
      endif
      et=1.0/e(a)
      ot=omem*a*et
      vt=omev*et*a**4
      graw=2.5*ot/(ot**(0.571428571)-vt+
     $     (1.0+0.5*ot)*(1.0+vt*0.014285714))
      gr=graw*grnorm
      return
      end
ccc      
      function e(a)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      common /omegas/ omem,omev,omek
      a2=a*a
      e=a*omem-a2*omek+a2*a2*omev
      return
      end

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccc      
ccc  package of solving the integral equation of 
ccc  affine parameter - redshit (scale factor) relation
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!!!
      function dw(a)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      common /cosmo/ om,ode,w0,wa
      ee=a*om+ode*a*a**(-3.0*(w0+wa))*exp(-3.0*wa*(1.0-a))
      dw=1.0/sqrt(ee)
      return
      end
!!!
      function dldz(a)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      common /omegas/ om,ov,ok
      a2=a*a
      dldz=1.0/sqrt(a*om-a2*ok+a2*a2*ov)      
      return
      end
cc

      SUBROUTINE qromb(func,a,b,ss)
      INTEGER JMAX,JMAXP,K,KM
      REAL a,b,func,ss,EPS
      EXTERNAL func
      PARAMETER (EPS=1.e-6, JMAX=40, JMAXP=JMAX+1, K=10, KM=K-1)
CU    USES polint,trapzd
      INTEGER j
      REAL dss,h(JMAXP),s(JMAXP)
      h(1)=1.
      do 11 j=1,JMAX
        call trapzd(func,a,b,s(j),j)
        if (j.ge.K) then
          call polint(h(j-KM),s(j-KM),K,0.,ss,dss)
          if (abs(dss).le.EPS*abs(ss)) return
        endif
        s(j+1)=s(j)
        h(j+1)=0.25*h(j)
11    continue
      pause 'too many steps in qromb'
      END
C  (C) Copr. 1986-92 Numerical Recipes Software 41m.
      SUBROUTINE polint(xa,ya,n,x,y,dy)
      INTEGER n,NMAX
      REAL dy,x,y,xa(n),ya(n)
      PARAMETER (NMAX=20)
      INTEGER i,m,ns
      REAL den,dif,dift,ho,hp,w,c(NMAX),d(NMAX)
      ns=1
      dif=abs(x-xa(1))
      do 11 i=1,n
        dift=abs(x-xa(i))
        if (dift.lt.dif) then
          ns=i
          dif=dift
        endif
        c(i)=ya(i)
        d(i)=ya(i)
11    continue
      y=ya(ns)
      ns=ns-1
      do 13 m=1,n-1
        do 12 i=1,n-m
          ho=xa(i)-x
          hp=xa(i+m)-x
          w=c(i+1)-d(i)
          den=ho-hp
          if(den.eq.0.)pause 'failure in polint'
          den=w/den
          d(i)=hp*den
          c(i)=ho*den
12      continue
        if (2*ns.lt.n-m)then
          dy=c(ns+1)
        else
          dy=d(ns)
          ns=ns-1
        endif
        y=y+dy
13    continue
      return
      END
C  (C) Copr. 1986-92 Numerical Recipes Software 41m.
      SUBROUTINE trapzd(func,a,b,s,n)
      INTEGER n
      REAL a,b,s,func
      EXTERNAL func
      INTEGER it,j
      REAL del,sum,tnm,x
      if (n.eq.1) then
        s=0.5*(b-a)*(func(a)+func(b))
      else
        it=2**(n-2)
        tnm=it
        del=(b-a)/tnm
        x=a+0.5*del
        sum=0.
        do 11 j=1,it
          sum=sum+func(x)
          x=x+del
11      continue
        s=0.5*(s+(b-a)*sum/tnm)
      endif
      return
      END
C  (C) Copr. 1986-92 Numerical Recipes Software 41m.

      SUBROUTINE hunt(xx,n,x,jlo)
      INTEGER jlo,n
      REAL x,xx(n)
      INTEGER inc,jhi,jm
      LOGICAL ascnd
      ascnd=xx(n).gt.xx(1)
      if(jlo.le.0.or.jlo.gt.n)then
        jlo=0
        jhi=n+1
        goto 3
      endif
      inc=1
      if(x.ge.xx(jlo).eqv.ascnd)then
1       jhi=jlo+inc
        if(jhi.gt.n)then
          jhi=n+1
        else if(x.ge.xx(jhi).eqv.ascnd)then
          jlo=jhi
          inc=inc+inc
          goto 1
        endif
      else
        jhi=jlo
2       jlo=jhi-inc
        if(jlo.lt.1)then
          jlo=0
        else if(x.lt.xx(jlo).eqv.ascnd)then
          jhi=jlo
          inc=inc+inc
          goto 2
        endif
      endif
3     if(jhi-jlo.eq.1)return
      jm=(jhi+jlo)/2
      if(x.gt.xx(jm).eqv.ascnd)then
        jlo=jm
      else
        jhi=jm
      endif
      goto 3
      END
C  (C) Copr. 1986-92 Numerical Recipes Software 41m.


      SUBROUTINE qromb2(func,a,b,ss)
      INTEGER JMAX,JMAXP,K,KM
      REAL a,b,func,ss,EPS
      EXTERNAL func
      PARAMETER (EPS=1.e-6, JMAX=40, JMAXP=JMAX+1, K=10, KM=K-1)
CU    USES polint2,trapzd2
      INTEGER j
      REAL dss,h(JMAXP),s(JMAXP)
      h(1)=1.
      do 11 j=1,JMAX
        call trapzd2(func,a,b,s(j),j)
        if (j.ge.K) then
          call polint2(h(j-KM),s(j-KM),K,0.,ss,dss)
          if (abs(dss).le.EPS*abs(ss)) return
        endif
        s(j+1)=s(j)
        h(j+1)=0.25*h(j)
11    continue
      pause 'too many steps in qromb'
      END
C  (C) Copr. 1986-92 Numerical Recipes Software 41m.
      SUBROUTINE polint2(xa,ya,n,x,y,dy)
      INTEGER n,NMAX
      REAL dy,x,y,xa(n),ya(n)
      PARAMETER (NMAX=20)
      INTEGER i,m,ns
      REAL den,dif,dift,ho,hp,w,c(NMAX),d(NMAX)
      ns=1
      dif=abs(x-xa(1))
      do 11 i=1,n
        dift=abs(x-xa(i))
        if (dift.lt.dif) then
          ns=i
          dif=dift
        endif
        c(i)=ya(i)
        d(i)=ya(i)
11    continue
      y=ya(ns)
      ns=ns-1
      do 13 m=1,n-1
        do 12 i=1,n-m
          ho=xa(i)-x
          hp=xa(i+m)-x
          w=c(i+1)-d(i)
          den=ho-hp
          if(den.eq.0.)pause 'failure in polint'
          den=w/den
          d(i)=hp*den
          c(i)=ho*den
12      continue
        if (2*ns.lt.n-m)then
          dy=c(ns+1)
        else
          dy=d(ns)
          ns=ns-1
        endif
        y=y+dy
13    continue
      return
      END
C  (C) Copr. 1986-92 Numerical Recipes Software 41m.
      SUBROUTINE trapzd2(func,a,b,s,n)
      INTEGER n
      REAL a,b,s,func
      EXTERNAL func
      INTEGER it,j
      REAL del,sum,tnm,x
      if (n.eq.1) then
        s=0.5*(b-a)*(func(a)+func(b))
      else
        it=2**(n-2)
        tnm=it
        del=(b-a)/tnm
        x=a+0.5*del
        sum=0.
        do 11 j=1,it
          sum=sum+func(x)
          x=x+del
11      continue
        s=0.5*(s+(b-a)*sum/tnm)
      endif
      return
      END
C  (C) Copr. 1986-92 Numerical Recipes Software 41m.

ccc spline fitting package
      SUBROUTINE splint(xa,ya,y2a,imin,imax,x,y)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      dimension xa(imin:imax)
      dimension y2a(imin:imax)
      dimension ya(imin:imax)
      klo=imin
      khi=imax
      zero=0.0
 1    if (khi-klo.gt.1) then
         k=(khi+klo)/2
         if(xa(k).gt.x)then
            khi=k
         else
            klo=k
         endif
         goto 1
      endif
      h=xa(khi)-xa(klo)
      if (h.eq.zero) pause 'bad xa input in splint'
      a=(xa(khi)-x)/h
      b=(x-xa(klo))/h
      y=a*ya(klo)+b*ya(khi)+((a**3-a)*y2a(klo)+
     $     (b**3-b)*y2a(khi))*(h**2)/6.0
      return
      END
c
      SUBROUTINE spline(x,y,imin,imax,yp1,ypn,y2)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      dimension x(imin:imax)
      dimension y(imin:imax)
      dimension y2(imin:imax)
      PARAMETER (iNMAX=5000)
      dimension u(-iNMAX:iNMAX)
      if (yp1.gt.0.99e30) then
        y2(imin)=0.0
        u(imin)=0.0
      else
         y2(imin)=-0.5
         u(imin)=(3.0/(x(imin+1)-x(imin)))*((y(imin+1)
     $        -y(imin))/(x(imin+1)-x(imin))-yp1)
      endif
      do 11 i=imin+1,imax-1
         sig=(x(i)-x(i-1))/(x(i+1)-x(i-1))
         p=sig*y2(i-1)+2.0
         y2(i)=(sig-1.)/p
         u(i)=(6.0*((y(i+1)-y(i))/(x(i+1)-x(i))-(y(i)-y(i-1))/
     $        (x(i)-x(i-1)))/(x(i+1)-x(i-1))-sig*u(i-1))/p
11    continue
      if (ypn.gt.0.99e30) then
         qn=0.0
         un=0.0
      else
         qn=0.5
         un=(3.0/(x(imax)-x(imax-1)))*(ypn-(y(imax)-y(imax-1))/
     $        (x(imax)-x(imax-1)))
      endif
      y2(imax)=(un-qn*u(imax-1))/(qn*y2(imax-1)+1.0)
      do 12 k=imax-1,imin,-1
         y2(k)=y2(k)*y2(k+1)+u(k)
12    continue
      return
      END
      
      FUNCTION gammln(xx)
      REAL gammln,xx
      INTEGER j
      DOUBLE PRECISION ser,stp,tmp,x,y,cof(6)
      SAVE cof,stp
      DATA cof,stp/76.18009172947146d0,-86.50532032941677d0,
     *24.01409824083091d0,-1.231739572450155d0,.1208650973866179d-2,
     *-.5395239384953d-5,2.5066282746310005d0/
      x=xx
      y=x
      tmp=x+5.5d0
      tmp=(x+0.5d0)*log(tmp)-tmp
      ser=1.000000000190015d0
      do 11 j=1,6
        y=y+1.d0
        ser=ser+cof(j)/y
11    continue
      gammln=tmp+log(stp*ser/x)
      return
      END
C  (C) Copr. 1986-92 Numerical Recipes Software 41m.
!!!
      function omegade(a)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      common /cosmo/ om,ode,w0,wa
      omegade=(1.0-om)/
     $     (1.0-om+om*a**(3.0*(w0+wa))*exp(3.0*wa*(1.0-a)))
      return
      end

      function omegam(a)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      common /cosmo/ om,ode,w0,wa
      omegam=om/(1.0-om+om*a**(3.0*(w0+wa))*exp(3.0*wa*(1.0-a)))/a**3
      return
      end

ccc linear growth factor
      function gra(a,nstep,hrk,dplus)
      implicit real (a-h,o-z)
      implicit integer (i-n)
      dimension dplus(0:nstep)
      x=a/hrk
      ix=int(x)
      dx=(x-ix)
      gra=dplus(ix)*(1.0-dx)+dplus(ix+1)*dx
      return
      end

      subroutine derivs(x,y,dy)
      implicit real (a-h,o-z) 
      implicit integer (i-n)
      common /cosmo/ om,ode,w0,wa

      parameter (n=2)
      real y(n),dy(n)
      
      a=x
      dy(1)=y(2)
      
      ww=w0+wa*(1.0-a)
      omede=omegade(a)

      dy(2)=-(3.5-1.5*ww*omede)*y(2)/a
     $     -1.5*(1.0-ww)*omede*y(1)/(a*a)
      
      return
      end
      
      SUBROUTINE rk4(y,dydx,n,x,h,yout,derivs)
      INTEGER n,NMAX
      REAL h,x,dydx(n),y(n),yout(n)
      EXTERNAL derivs
      PARAMETER (NMAX=50)
      INTEGER i
      REAL h6,hh,xh,dym(NMAX),dyt(NMAX),yt(NMAX)
      hh=h*0.5
      h6=h/6.
      xh=x+hh
      do 11 i=1,n
        yt(i)=y(i)+hh*dydx(i)
11    continue
      call derivs(xh,yt,dyt)
      do 12 i=1,n
        yt(i)=y(i)+hh*dyt(i)
12    continue
      call derivs(xh,yt,dym)
      do 13 i=1,n
        yt(i)=y(i)+h*dym(i)
        dym(i)=dyt(i)+dym(i)
13    continue
      call derivs(x+h,yt,dyt)
      do 14 i=1,n
        yout(i)=y(i)+h6*(dydx(i)+dyt(i)+2.*dym(i))
14    continue
      return
      END
C  (C) Copr. 1986-92 Numerical Recipes Software 41m.
!!!
